#include <stdio.h>
#include <stdlib.h>
#include "CTurtle.h"

enum out_format_t {INVALID=-1, TARGA, ASCII_PPM, BINARY_PPM};

void hilbert_l(Image img, int lvl, int w);
void hilbert_r(Image img, int lvl, int w);

// "left hand" side of the hilbert curve algorithm
void hilbert_l(Image img, int lvl, int w) {
    if (lvl==0) return;
    turtle_right(img,90);
    hilbert_r(img, lvl-1, w);
    turtle_forward(img, w);
    turtle_left(img,90);
    hilbert_l(img, lvl-1, w);
    turtle_forward(img, w);
    hilbert_l(img, lvl-1, w);
    turtle_left(img,90);
    turtle_forward(img, w);
    hilbert_r(img, lvl-1, w);
    turtle_right(img,90);
}

// "right hand" side of the hilbert curve algorithm
void hilbert_r(Image img, int lvl, int w) {
    if (lvl==0) return;
    turtle_left(img,90);
    hilbert_l(img, lvl-1, w);
    turtle_forward(img, w);
    turtle_right(img,90);
    hilbert_r(img, lvl-1, w);
    turtle_forward(img, w);
    hilbert_r(img, lvl-1, w);
    turtle_right(img,90);
    turtle_forward(img, w);
    hilbert_l(img, lvl-1, w);
    turtle_left(img,90);
}

// draw a hilbert curve with the given complexity lvl and segment length w
void hilbert(Image img, int lvl, int w) {
    hilbert_l(img, lvl, w);
}

// print usage and exit with status 1, optionally displaying an error message if provided one
void usage(char* msg) {
    printf("CTurtle test suite for C.\nSyntax:\n  ./cturtle_test <test#> <outfile> [infile]\n\n");
    if (msg) {
        printf("\nError: %s\n",msg);
    }
    exit(1);
}

// write a ppm, and report errors
void verbose_write_ppm(const char* filename, Image img, bool ascii) {
    if (!write_ppm(filename, img, ascii)) {
        printf("%s: %s\n",filename, get_last_error());
    }
}

// write a tga, and say we're doing it
void verbose_write_tga(const char* filename, Image img) {
    if (!write_tga(filename, img)) {
        printf("%s: %s\n",filename, get_last_error());
    }
}

// basic test image - a 3x5 image with r g b dots going down the middle and the four corner pixels set to white.
// used for multiple tests
Image make_simple_base_image() {
    int w=3;
    int h=5;
    Image img = create_image(w,h);
    
    // r g b dots going down the middle
    do_pixel(img,1,0,(Color) {255,0,0});
    do_pixel(img,1,2,(Color) {0,255,0});
    do_pixel(img,1,4,(Color) {0,0,255});
    
    // 4 corners in white
    do_pixel(img,0  ,0  ,(Color) {255,255,255});
    do_pixel(img,0  ,h-1,(Color) {255,255,255});
    do_pixel(img,w-1,0  ,(Color) {255,255,255});
    do_pixel(img,w-1,h-1,(Color) {255,255,255});
    
    return img;
    
}

// write the given image to the test-out directory in all 3 formats
void write_test_image_outputs(Image img, const char* outfile_base) {
    char filename[64];
    sprintf(filename,"%sA.ppm",outfile_base); verbose_write_ppm(filename,img,true);
    sprintf(filename,"%sB.ppm",outfile_base); verbose_write_ppm(filename,img,false);
    sprintf(filename,"%s.tga",  outfile_base); verbose_write_tga(filename,img);
}


bool ends_with(const char* string, const char* substr) {
    const char* a = string + strlen(string);
    const char* b = substr + strlen(substr);
    while (1) {
        if (*a-- != *b--) return false;
        if (b<substr)     return true;  // got all the way through substr with all matches; true!
        if (a<string)     return false; // ran out of string before substr, string is shorter than substr, false!
    }
    return true; // we should never get here
}

enum out_format_t get_format(const char* filename) {
    if      (ends_with(filename,   ".tga")) return TARGA;
    else if (ends_with(filename, ".a.ppm")) return ASCII_PPM;
    else if (ends_with(filename,   ".ppm")) return BINARY_PPM;
    else return INVALID;
}

// write the given image to the given filename in the given format
void write_output(Image img, const char* out_filename) {
    switch (get_format(out_filename)) {
        case TARGA:      verbose_write_tga(out_filename, img);        break;
        case ASCII_PPM:  verbose_write_ppm(out_filename, img, true);  break;
        case BINARY_PPM: verbose_write_ppm(out_filename, img, false); break;
        default:
            printf("Tried to write an invalid format!\n");
            exit(1);
    }
}


Image read_image(const char* filename) {
    Image img;
    if      (ends_with(filename,".tga"))  img = read_tga(filename);
    else if (ends_with(filename,".ppm"))  img = read_ppm(filename);
    else {
        printf("Error: %s doesn't have an appropriate extension.\n",filename);
        exit(1);
    }
    
    if (!img) {
        printf("%s: %s\n", filename, get_last_error());
        exit(1);
    }
    
    return img;
}

// run a numbered test and save results
int main(int argc, char *argv[]) {
    if (argc < 3) usage(NULL);
    
    int test_id;
    if (!sscanf(argv[1],"%d",&test_id)) usage("Unable to parse id");
    
    const char* out_filename = argv[2];
    enum out_format_t out_format = get_format(out_filename);
    if (out_format == INVALID) {
        printf("Invalid output image format.\n");
        exit(1);
    }
    
    const char* input_filename = NULL; // optional
    if (argc >= 4) input_filename = argv[3];
    
    // vars used in the tests
    int w,h; 
    int i;
    Image img;
    
    // for two-output tests:
    Image img1;
    Image img2;
    char out_filename1[strlen(out_filename)+1];
    char out_filename2[strlen(out_filename)+1];
    
    printf("Running test %d (%s)...\n",test_id, out_filename);
    
    switch (test_id) {
    case 1:
        // basic test image - a 3x5 image with r g b dots going down the middle and the four corner pixels set to white.
        w=3; h=5;
        img = make_simple_base_image();
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 2:
        // same as test 1, but invert at the end
        img = make_simple_base_image();
        
        do_invert(img);

        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 3:
        // same as test 1, but reverse rgb at the end
        img = make_simple_base_image();
        
        do_reverse_rgb(img);

        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 4:
        w=150; h=60;
        img = create_image(w,h);
        
        // this line will be covered up by the clear operation and shouldn't be seen
        do_line(img, 0,0, w-1, h-1, (Color) {255,255,255}); // white
        
        // clear the canvas
        do_clear(img, (Color) {0,0,40}); // dark blue background
        
        // upper triangle
        do_line(img, 50,20, 60,30, (Color) {255,000,000}); // red
        do_line(img, 50,20, 40,30, (Color) {000,255,000}); // green
        do_line(img, 40,30, 60,30, (Color) {000,000,255}); // blue

        // lower left triangle
        do_line(img, 50,40, 60,30, (Color) {255,000,255}); // magenta
        do_line(img, 50,40, 70,40, (Color) {000,255,255}); // cyan
        do_line(img, 70,40, 60,30, (Color) {255,255,000}); // yellow

        // lower right triangle
        do_line(img, 50,40, 40,30, (Color) {128,000,000}); // dk red
        do_line(img, 50,40, 30,40, (Color) {000,128,000}); // dk green
        do_line(img, 30,40, 40,30, (Color) {000,000,128}); // dk blue

        // vertical lines on the sides
        do_line(img, 20,10, 20,50, (Color) {128,128,128}); // grey
        do_line(img, 80,10, 80,50, (Color) {128,128,128}); // grey
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 5:
        // read the given file (auto-detecting type by extension) and output it as the test output files (no changes)
        if (!input_filename) usage("Need input filename for this test");
        img = read_image(input_filename);
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 6:
        // read the given file (auto-detecting type by extension) and output it with rgb reversed
        if (!input_filename) usage("Need input filename for this test");
        img = read_image(input_filename);
        
        do_reverse_rgb(img);
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 7:
        // read the given file (auto-detecting type by extension) and invert it, then draw most of test 2 on top of it
        if (!input_filename) usage("Need input filename for this test");
        img = read_image(input_filename);
        
        do_invert(img);
        
        // upper triangle
        do_line(img, 50,20, 60,30, (Color) {255,000,000}); // red
        do_line(img, 50,20, 40,30, (Color) {000,255,000}); // green
        do_line(img, 40,30, 60,30, (Color) {000,000,255}); // blue

        // lower left triangle
        do_line(img, 50,40, 60,30, (Color) {255,000,255}); // magenta
        do_line(img, 50,40, 70,40, (Color) {000,255,255}); // cyan
        do_line(img, 70,40, 60,30, (Color) {255,255,000}); // yellow

        // lower right triangle
        do_line(img, 50,40, 40,30, (Color) {128,000,000}); // dk red
        do_line(img, 50,40, 30,40, (Color) {000,128,000}); // dk green
        do_line(img, 30,40, 40,30, (Color) {000,000,128}); // dk blue

        // vertical lines on the sides
        do_line(img, 20,10, 20,50, (Color) {128,128,128}); // grey
        do_line(img, 80,10, 80,50, (Color) {128,128,128}); // grey
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
        
    case 8:
        // the "hello world" of turtle graphics -- drawing one line with the default settings in the default direction
        w=50; h=50;
        img = create_image(w,h);
        
        turtle_forward(img,20);
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 9:
        // same as 8, but turn a bit first
        w=50; h=50;
        img = create_image(w,h);
        
        turtle_right(img,30);
        turtle_forward(img,20);
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 10:
        // simple triangle with different colors on each edge (red/orange/yellow)
        w=50; h=50;
        img = create_image(w,h);
        
        turtle_set_xy(img, 5,5);
        
        for (i=0; i<3; i++) {
            turtle_set_color(img, (Color){255,i*120,0});
            turtle_right(img, 120);
            turtle_forward(img,30);
        }
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 11:
        // draw a magenta box with a cyan X on it a bit off center
        w=60; h=45;
        img = create_image(w,h);
        
        turtle_set_color(img, (Color){255,0,255});
        turtle_set_xy(img, 10,10);
        turtle_move(img,   40,10);
        turtle_move(img,   40,40);
        turtle_move(img,   10,40);
        turtle_move(img,   10,10);
        
        turtle_set_color(img, (Color){0,255,255});
        turtle_move(img,   40,40);
        turtle_set_xy(img, 40,10);
        turtle_move(img,   10,40);
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 12:
        // draw a "zipper" type shape going from blue to cyan
        w=50; h=80;
        img = create_image(w,h);
        
        turtle_set_xy(img, w/2,h-5);
        for (i=0; i<4; i++) {
            turtle_set_color(img, (Color){0,85*i,255});
            
            turtle_face_north(img);
            turtle_forward(img, 10);
            
            turtle_left(img, 100);
            turtle_forward(img, 10);
            
            turtle_face_north(img);
            turtle_forward(img, 10);
            
            turtle_right(img, 100);
            turtle_forward(img, 10);
        }
            
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 13:
        // draw a dashed hexagon
        w=170; h=200;
        img = create_image(w,h);
        
        int n=6;
        turtle_set_xy(img, 3,140);
        for (i=0; i<n; i++) {
            
            for (int j=0; j<3; j++) {
                turtle_pen_up(img);
                turtle_forward(img, 10);
                turtle_pen_down(img);
                turtle_forward(img, 20);
            }
            turtle_right(img, 360/n);
            
        }
            
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 14:
        // draw a totally sweet hilbert curve fractal
        w=256; h=256;
        img = create_image(w,h);
        
        do_clear(img, (Color){15,0,30});
        
        turtle_set_color(img, (Color){255,128,0});
        
        turtle_set_xy(img, 1,h-2);
        hilbert(img,6,4);
        
        write_output(img, out_filename);
        destroy_image(img);
        
        break;
        
    case 15:
        // make two images at once
        if (!strstr(out_filename,"%d")) {
            printf("This test must have an output filename containing '%%d', as it outputs multiple numbered files at once.\n");
            exit(1);
        }
        w=70; h=60;
        img1 = create_image(w,h);
        img2 = create_image(w+5,h+9);
        
        do_clear(img1, (Color){15,0,30});
        do_clear(img2, (Color){45,2,4});
        
        turtle_set_color(img1, (Color){255,128,0});
        turtle_set_color(img2, (Color){0,64,255});
        
        turtle_set_xy(img1, w/2-9,  h/2-5);
        turtle_set_xy(img2, w/2-12, h/2-17);
        
        turtle_forward(img1,  13);
        turtle_forward(img2, -33);
        
        turtle_right(img1, 30);
        turtle_left (img2, 45);
        
        turtle_forward(img1,  11);
        turtle_forward(img2, -19);
        
        sprintf(out_filename1,out_filename,1);
        sprintf(out_filename2,out_filename,2);
        
        write_output(img1, out_filename1);
        write_output(img2, out_filename2);
        
        destroy_image(img1);
        destroy_image(img2);
        
        break;
        
    default:
        usage("Invalid test id");
    }
    
    printf("OK.\n");
    return EXIT_SUCCESS;
    
}
