#!/bin/bash
FAIL=0

VERBOSE=0

make clean
make
if [ $? -ne 0 ]; then
  echo "**** Make (compilation) FAILED"
  FAIL=1
fi

test_report() {
    local TEST_ID=$1
    local RESULT=$2
    [ $VERBOSE -ne 0 ] && echo -ne "\e[37;1;44m"
    printf "Test %-40s %s" "$TEST_ID" "$RESULT"
    [ $VERBOSE -ne 0 ] && echo -ne "\e[m"
    echo ""
}

run_basic_test() {
    local TEST_ALG=$1
    local EXP_RETVAL=$2
    local INPUT_FILE=$3 # optional
    local CLIENT=$4
    local TEST_ID=$5
    local NUM_OUTPUTS=$6 # usually 1, except for a multi-output test case like T15

    ACT_LOG="test-out/actual-${TEST_ID}.log"
    EXP_LOG="test-out/expected-${TEST_ID}.log"

    ACT_OUTFILE="test-out/actual-${TEST_ID}"
    EXP_OUTFILE="test-out/expected-${TEST_ID}"

    if [ "$NUM_OUTPUTS" -gt 1 ] ; then
        ACT_OUTFILE=${ACT_OUTFILE/-c./-%d-c.}
        ACT_OUTFILE=${ACT_OUTFILE/-py./-%d-py.}
        EXP_OUTFILE=${EXP_OUTFILE/-c./-%d-c.}
        EXP_OUTFILE=${EXP_OUTFILE/-py./-%d-py.}
    fi

    EXP_OUTFILE=${EXP_OUTFILE/-c./.}
    EXP_OUTFILE=${EXP_OUTFILE/-py./.}

    rm -f $ACT_LOG $ACT_OUTFILE

    COMMAND="$CLIENT $TEST_ALG $ACT_OUTFILE $INPUT_FILE"

    [ $VERBOSE -ne 0 ] && echo -e "\e[33m"\$ $COMMAND "\e[m"
    $COMMAND > $ACT_LOG 2>&1
    local ACT_RETVAL=$?
    [ $VERBOSE -ne 0 ] && ( echo -ne "\e[36m" ; cat $ACT_LOG ; echo -ne "\e[m" )

    local LOCALFAIL=0

    if [ $ACT_RETVAL -ne $EXP_RETVAL ]; then
        test_report "$TEST_ID" "FAILED - RETVAL wrong. Expected: $EXP_RETVAL Actual: $ACT_RETVAL"
        LOCALFAIL=1
        FAIL=1
    fi

    if (! diff -q $EXP_LOG $ACT_LOG ); then
        test_report "$TEST_ID" "FAILED - console output did not match"
        LOCALFAIL=1
        FAIL=1
    fi

    if [ "$NUM_OUTPUTS" -ne 1 ] ; then
        for I in `seq 1 $NUM_OUTPUTS` ; do
            THIS_ACT_OUTFILE=`printf "$ACT_OUTFILE" $I`
            THIS_EXP_OUTFILE=`printf "$EXP_OUTFILE" $I`
            if (! diff -q $THIS_ACT_OUTFILE $THIS_EXP_OUTFILE ); then
                test_report "$TEST_ID" "FAILED - image output did not match"
                LOCALFAIL=1
                FAIL=1
            fi
        done
    else
        if (! diff -q $ACT_OUTFILE $EXP_OUTFILE ); then
            test_report "$TEST_ID" "FAILED - image output did not match"
            LOCALFAIL=1
            FAIL=1
        fi
    fi


    if [ $LOCALFAIL -eq 0 ]; then
        test_report "$TEST_ID" "PASS"
        #printf "Test %-15s %s" "$TEST_ID" "PASS"
    fi

    echo ""

    return $FAIL
}


run_test() {
    local TEST_ALG=$1
    local EXP_RETVAL=$2
    local NUM_OUTPUTS=$3 # usually 1, except for a multi-output test case like T15
    local INPUT_FILE=$4 # optional

    TEST_ID_BASE="T${TEST_ALG}"
    if [ -n "$INPUT_FILE" ] ; then TEST_ID_BASE+="-"`basename $INPUT_FILE` ; fi

    run_basic_test $TEST_ALG $EXP_RETVAL "$INPUT_FILE" "./cturtle_test"     "${TEST_ID_BASE}-c.tga" $NUM_OUTPUTS
    run_basic_test $TEST_ALG $EXP_RETVAL "$INPUT_FILE" "./pyturtle_test.py" "${TEST_ID_BASE}-py.tga" $NUM_OUTPUTS
    [ "$NUM_OUTPUTS" -eq 0 ] && return   # 0-output tests only have to run once each for c/python, since they're just invalid read tests

    run_basic_test $TEST_ALG $EXP_RETVAL "$INPUT_FILE" "./cturtle_test"     "${TEST_ID_BASE}-c.a.ppm" $NUM_OUTPUTS
    run_basic_test $TEST_ALG $EXP_RETVAL "$INPUT_FILE" "./pyturtle_test.py" "${TEST_ID_BASE}-py.a.ppm" $NUM_OUTPUTS

    run_basic_test $TEST_ALG $EXP_RETVAL "$INPUT_FILE" "./cturtle_test"     "${TEST_ID_BASE}-c.b.ppm" $NUM_OUTPUTS
    run_basic_test $TEST_ALG $EXP_RETVAL "$INPUT_FILE" "./pyturtle_test.py" "${TEST_ID_BASE}-py.b.ppm" $NUM_OUTPUTS

    return

}

echo ""
echo "Running system tests..."
echo ""

# each run_test below tests all 3 output formats and both C and Python clients
# (unless NUM_OUTPUTS is 0, in which case it only tries tga output, since that indicates a test of bad input files, so no image should get written anyway)
# Parameters: 
# TEST_ALG EXP_RETVAL NUM_OUTPUTS INPUT_FILE
run_test 01 0 1 ""
run_test 02 0 1 ""
run_test 03 0 1 ""
run_test 04 0 1 ""
run_test 05 0 1 "test-in/toucan.tga"
run_test 05 0 1 "test-in/toucan.a.ppm"
run_test 05 0 1 "test-in/toucan.b.ppm"
run_test 05 0 1 "test-in/scaled.a.ppm"
run_test 05 1 0 "test-in/pgm_unsupp_grey.ppm"
run_test 05 1 0 "test-in/tga_unsupp_colormap.tga"
run_test 05 1 0 "test-in/tga_unsupp_compression.tga"
run_test 05 1 0 "test-in/tga_unsupp_image_desc.tga"
run_test 06 0 1 "test-in/toucan.a.ppm"
run_test 07 0 1 "test-in/toucan.a.ppm"
run_test 08 0 1 ""
run_test 09 0 1 ""
run_test 10 0 1 ""
run_test 11 0 1 ""
run_test 12 0 1 ""
run_test 13 0 1 ""
run_test 14 0 1 ""
run_test 15 0 2 ""

if [ $FAIL -eq 1 ]; then
  echo "FAILING TESTS!"
  exit 13
else
  exit 0
fi

